package PureMessage::Policy::Delay::Rescan;

use strict;
use warnings;

use IO::Socket::INET;
use IO::Select;
use File::Basename;

use PureMessage::MessageFile;
use PureMessage::Logger qw(LOG_ERR LOG_INFO LOG_DEBUG);
use POSIX qw(strftime);
use base 'Exporter';
use Math::Round;
use PureMessage::Version qw($PREFIX);
use PureMessage::_sys ();
use PureMessage::Config;

our @EXPORT_OK = qw(rescan_messages);

our %config;
PureMessage::Config->import('%config');

our $RESCAN_DIR="$PREFIX/var/qdir/delay/rescan";

my $peeraddr;
my $default_log_file = "$PREFIX/var/log/delay.log";
my $localhost = PureMessage::_sys::gethostname();
my $last_mod_time_delay = 0;
my $last_mod_time_pmx = 0;
my $delay_conf_file = "$PREFIX/etc/delay.conf";
my $pmx_conf_file = "$PREFIX/etc/pmx.conf.cache";
my $debug_level = 0;

my $logger = PureMessage::Logger->new(
    ident => 'delay_queue_runner',
    log_to => $default_log_file
);

sub _get_response {
    my $sock = shift || return;
    my $excepted = shift || '2..';
    my $noselect = shift;
    my $buf = shift;
    my $retcode;

    if (!$noselect) {
        my $rset = new IO::Select();
        $rset->add($sock);
        my ($rh_set) = IO::Select->select($rset, undef, undef, 10);
        return if (!$rh_set);
    }

    while(!defined $retcode) {
        my $line = $sock->getline() || return;
	${$buf} .= $line if ($buf);
	if (substr($line,3,1) eq ' ') {
	    $retcode = substr($line,0,3);
	    last;
	}
    }

    return ($retcode =~ /^$excepted$/);
}

sub _send_cmd {
    my $sock = shift || return;
    my $cmd = shift || return;
    my $excepted = shift;
    $sock->print($cmd."\r\n") || return;
    return _get_response($sock, $excepted);
}


sub _send_data {
    my $sock = shift || return;
    my $st = shift || return;
    my $message = $st->{file} || return;
    my $size = $st->{mf}->message_size || return;
    my $write_err = 0;
    my $read = 0;
    my $eoh = 0;

    my ($reason,$header);
    if ($message =~ /_([frt])_([0-9]{1,2})$/) {
	$reason = "rr=".(($1 eq 'r') ? 'release' : ($1 eq 'f') ? 'full' : 'timeout').",bd=$2";
    } else {
	$reason = "rr=timeout,bd=0";
    }

    open(my $fh, '<', $message) || return;
    if (!_send_cmd($sock, "DATA", '354')) {
	close $fh;
	return;
    }
    while (<$fh>) {
	$read += length($_);
	if ($eoh) {
	    $_ = ".$_" if (/^\.+\r?$/);
	} else {

	    # skip X-Sophos-ReScanReason header for now, we will add this at eoh
	    if ( /^X-Sophos-ReScanReason: /){
		last if ($read >= $size);
		next;
            }

	    # skip X-Sophos-ReScan header , it will be added by milter
	    if ( /^X-Sophos-ReScan: /){
		last if ($read >= $size);
		next;
            }

	    if ( /^\r?$/ ) {

		if (!$sock->print("X-Sophos-ReScanReason: $reason\r\n")) {
		    $write_err = 1;
		    last;
		}
		$eoh = 1;
	    }
	}
	if (!$sock->print($_)) {
	    $write_err = 1;
	    last;
	}
	last if ($read >= $size);
    }
    close $fh;

    return if ($write_err);
    $sock->print(".\r\n");

    return 1;
}

sub _send_mail {
    my $sock = shift || return;
    my $st = shift || return;
    my $ret = 0;

    my $mf = $st->{mf} || return;
    my $from = $mf->minfo->value('Envelope-From');

    if ($st->{helo} == 0) {
       # my $helo = $mf->minfo->value('Helo') || '';
        if (_send_cmd($sock, "HELO $localhost")) {
            _send_cmd($sock, 'RESCAN');
            $st->{helo} = 1;
        }
        return if ($st->{helo} == 0);
    }

    if (defined $from and _send_cmd($sock, "MAIL FROM: <$from>")) {
        my $rcpts = 0;
        for my $to ($mf->minfo->value('Envelope-To')) {
            if (!_send_cmd($sock, "RCPT TO: <$to>")) {
                $rcpts = 0;
                last;
            }
            $rcpts++;
        }
	$st->{'Queue-Id'} = 0;
	$st->{'Queue-Id'} = $mf->minfo->value('Queue-Id');
        $ret = _send_data($sock, $st) if ($rcpts);
    }

    return $ret;
}

sub _close_socket {
    my $stat = shift || return;
    my $sock = shift || return;
    my $rset = shift || return;
    my $noquit = shift;

    $rset->remove($sock);
    _send_cmd($sock, "QUIT") if (!$noquit);
    delete $stat->{$sock};
    close $sock;

    return;
}

sub _create_socket {
    my $try = shift || 10;

    my $sock;

    while (!$sock && $try) {
        $sock = IO::Socket::INET->new(PeerAddr => $peeraddr, Timeout => 10);
        sleep 1 if (!$sock && --$try);
    }
    if (!$sock) {
	$logger->log(LOG_ERR, "not able to connect to $peeraddr\n");
    }
    return $sock;
}

sub _create_sockets {
    my $stat = shift || return;
    my $rset = shift || return;
    my $maxsockets = shift || return;
    my $filenum = shift || return;

    my $socketnum = scalar(keys %{$stat});
    my $created = 0;

    while ($socketnum + $created < $maxsockets and $filenum) {
        my $sock = _create_socket() || last;
        $rset->add($sock);
        my %s = ( helo => 0, data => 0, sock => \$sock, count => 0);
        $stat->{$sock} = \%s;
        last if(++$created >= $filenum);
    }
}


sub refresh_config {
    if (-e $delay_conf_file) {
	my $currtime = (stat ($delay_conf_file))[9];
	if ($last_mod_time_delay != $currtime) {
	    PureMessage::Policy::Delay::Config::refresh();
	    $debug_level = $PureMessage::Policy::Delay::Config::config{debug_level};
	    $last_mod_time_delay = $currtime;
	    $peeraddr = $PureMessage::Policy::Delay::Config::config{send_delayed_mail_to};
	    if (!$peeraddr) {
		$logger->log(LOG_ERR, "$PREFIX/etc/delay.conf file has not send_delayed_mail_to parameter\n");
	    }
	}
    }
    if (-e $pmx_conf_file) {
	my $currtime = (stat ($pmx_conf_file))[9];
	if ($last_mod_time_pmx != $currtime) {
	    $last_mod_time_pmx = $currtime;
	    PureMessage::Config->import('%config');
	}
    }
}


sub rescan_messages {
    my (@files) = @_;
    my $sockets = 1;
    my $milters;

    refresh_config();
    $milters = $config{milter}{Policy}{max_pooled_procs} || 1;
    $sockets = int($milters / 2) if ($milters > 1);

    my $rset = new IO::Select();
    my %stat;

    my $count = 0;

    _create_sockets(\%stat, $rset, $sockets, scalar(@files));
    while ($rset->count) {
	my ($rh_set) = IO::Select->select($rset, undef, undef, 300);

        last if (!$rh_set);
        for my $sock (@$rh_set) {
            my $st = $stat{$sock} || next;
            if ($st->{data} == 0) {
                # Read banner
                if (!_get_response($sock, '2.0', 1)) {
                    _close_socket(\%stat, $sock, $rset);
                    next;
                }
                my $file = shift(@files);
                if ($file) {
                    $file = "$RESCAN_DIR/$file";
                    if (-f $file) {
                        eval {
                            my $mf = PureMessage::MessageFile->new($file);
                            if ($mf) {
                                $st->{mf} = $mf;
                                $st->{file} = $file;
                                if (_send_mail($sock, $st)) {
                                    $st->{data} = 1;
                                }
                            }
                        };
                        if ($@) {
			    $logger->log(LOG_ERR, "message sending failed: $@\n");
                        }
                    }
                }
                _close_socket(\%stat, $sock, $rset) if (!$st->{data});
            } elsif ($st->{data}) {
                # Read DATA response
                my $buf;
                if (_get_response($sock, '250', 1, \$buf)) {
		    if ( $st->{file} ) {
			my $base = basename( $st->{file});
		    	$logger->log(LOG_ERR, "delay file=$base for Queue-Id=$st->{'Queue-Id'} is sent for rescan") if $debug_level > 0;
		    }
		    $logger->log(LOG_ERR, "Queue-Id=$st->{'Queue-Id'} successfully sent for rescan");
                    if ($st->{file}) {
                        $count++;
                        unlink $st->{file};
                    }
                }
                my $cont = 0;
                if (++$st->{count} < 10 && scalar(@files)) {
                    $st->{data} = 0;
                    $cont = $sock->print("NOOP\r\n") ;
                }
                _close_socket(\%stat, $sock, $rset) if (!$cont);
            } else {
                last;
            }
        }
        _create_sockets(\%stat, $rset, $sockets, scalar(@files));
    }

    for my $sock (keys %stat) {
        _close_socket(\%stat, ${$stat{$sock}->{sock}}, $rset, 1);
    }

    return $count;
}

1;
