package PureMessage::Policy::Delay::Runner;
use strict;
use warnings;

use File::Find;
use File::Copy;

use Sophos::Config qw(:path);
use PureMessage::Logger qw(LOG_ERR LOG_DEBUG LOG_INFO);
use PureMessage::Version qw($PREFIX);
use PureMessage::Policy::Delay::DelayqPmdb;
use PureMessage::Policy::Delay::DelayRetry;
use PureMessage::Policy::Delay::Config qw(%config);
use Data::Dumper;

my $qdirpath = "$PREFIX/var/qdir";
my $delay_conf_file = "$PREFIX/etc/delay.conf";
my $debug_level = 0;


our @ISA = qw( Exporter );
our @EXPORT_OK = qw(
    check_queue_size set_delay_queue_size
    release_next_mails rescan_next_mails
    process_ui_action
  
    $work_directory $rescan_directory $current_directory
    $max_queue_dir_size $release_percent $db_entry_directory $location_id
);

our @EXPORT = qw(
    check_queue_size set_delay_queue_size
    release_next_mails rescan_next_mails
    process_ui_action
  
    $work_directory $rescan_directory $current_directory
    $max_queue_dir_size $release_percent $db_entry_directory $location_id
);

our ( $work_directory, $rescan_directory, $current_directory,
      $max_queue_dir_size, $release_percent, $location_id,
      $delay_queue_act_size, $size_to_free, $db_entry_directory);

my $default_log_file = "$PREFIX/var/log/delay.log";
my $last_mod_time = 0;
my $logger = PureMessage::Logger->new(
    ident => 'delay_queue_runner',
    log_to => $default_log_file
);

sub refresh_config {
    if (-e $delay_conf_file) {
	my $currtime = (stat ($delay_conf_file))[9];
	if ($last_mod_time != $currtime) {
	    PureMessage::Policy::Delay::Config::refresh();
	    $last_mod_time = $currtime;

	    $max_queue_dir_size = $config{delay_max_size};
	    if (!$max_queue_dir_size) {
		$max_queue_dir_size = 1024;
	    }
	    $max_queue_dir_size = $max_queue_dir_size*1024*1024;
	    $debug_level = $config{debug_level};
	}
    }
}

sub check_queue_size() {
    my $queue_full;
    my $delay_db_conf = Sophos::Config->new;
    $delay_db_conf->parse_file("$work_directory/delay_queue_size");
    $delay_queue_act_size = path_value($delay_db_conf->root, "disk_used");


    if ( $queue_full = ($delay_queue_act_size >= $max_queue_dir_size) ) {
	$size_to_free  = $max_queue_dir_size * $release_percent;
	$size_to_free += $delay_queue_act_size - $max_queue_dir_size;
    }

    return $queue_full;
}

sub set_delay_queue_size {
    my ( $size_to_release, $messages_to_release ) = @_;
    return unless $size_to_release;

    my $delay_db_conf = Sophos::Config->new;
    $delay_db_conf->parse_file(
	"$work_directory/delay_queue_size",
	create   => 1,
	writable => 1
    );

    path_value(
	$delay_db_conf->root, "disk_used",
	path_value($delay_db_conf->root, "disk_used") - $size_to_release
    );

    path_value(
      $delay_db_conf->root, "message_count",
      path_value($delay_db_conf->root, "message_count") - $messages_to_release
    );

    $delay_db_conf->save;
}

sub release_next_mails {
    my $min_original        = shift;
    my $min                 = $min_original;
    my $size_to_release     = 0;
    my $messages_to_release = 0;
    my $before_delay        = 0;
    my $filesize;
    my @gid;
    my $gid;

    do {
	$before_delay = $min >= $min_original ?
	    $min - $min_original : 60 + $min - $min_original;

	$current_directory = "$work_directory/". $min++;
	$min = 0 if $min == 60;

	find(sub {
	    if ( -f $_ && $size_to_release < $size_to_free ) {
		$filesize = -s "$current_directory/$_";
		if ( move("$current_directory/$_", "$rescan_directory/$_\_f_$before_delay") ) {
		    $gid = "$_-$location_id";
		    push @gid, $gid;
		    ++$messages_to_release;
		    $size_to_release += $filesize;
		    $logger->log(LOG_ERR, "delay file=$_ gid=$gid is marked for rescan") if $debug_level > 0;
		} 
	    }
 	}, $current_directory);
    } while ( $size_to_release < $size_to_free );

    return if !$size_to_release;
    set_delay_queue_size($size_to_release, $messages_to_release);

    #remove these records from db.
    my %opts=('logger'=>$logger);
    eval {
	my $pmdb= PureMessage::Policy::Delay::DelayqPmdb->new($qdirpath, %opts);
	$pmdb->_trash_delayq_messages(\@gid);
    };
    if ($@) {
	$logger->log(LOG_ERR, "_trash_delayq_messages failed for gid @gid, db can not be updated. err [$@]");
    }

}

sub rescan_next_mails {
    my $size_to_release     = 0;
    my $messages_to_release = 0;
    my $filename;
    my $filesize;
    my @gid;
    my $gid;

    opendir(CURRENT_DIR, $current_directory);
    foreach (readdir(CURRENT_DIR)) {
	next if $_ =~ m/^\./;
	
	$gid = "$_-$location_id";
	$filename = "$current_directory/$_";
	$filesize = -s "$filename";
	if( -f $filename && move("$filename", $rescan_directory) )  {
	    ++$messages_to_release;
	    $size_to_release += $filesize;
	    $logger->log(LOG_ERR, "delay file=$_  gid=$gid is marked for rescan") if $debug_level > 0;
	    #prepare list of gid to be removed;
	    push @gid, $gid;
	}
    }

    closedir(CURRENT_DIR);

    return if !$size_to_release;
    set_delay_queue_size($size_to_release, $messages_to_release);

    #remove these records from db.
    my %opts=('logger'=>$logger);
    eval {
	my $pmdb= PureMessage::Policy::Delay::DelayqPmdb->new($qdirpath, %opts);
	$pmdb->_trash_delayq_messages(\@gid);
    };
    if ($@) {
	$logger->log(LOG_ERR, "_trash_delayq_messages failed for gid @gid, db can not be updated. err[$@]");
    }
}

sub process_ui_action {
    my %opts=('logger'=>$logger);
    my $retry_entries;
    my $delete_entries;
    
    refresh_config();
    eval {
	my $pmdb=   PureMessage::Policy::Delay::DelayqPmdb->new($qdirpath, %opts);
	$delete_entries = $pmdb->get_message_hash('delete'); 
	$retry_entries = $pmdb->get_message_hash('retry');
    };
    if ($@) {
	$logger->log(LOG_ERR, "DelayqPmdb.pm get_message_hash failed, can't process ui action if any. err[$@]");
	return;
    } 

    if (%{$delete_entries}) {
	$logger->log(LOG_ERR, "entries to be deleted:".Dumper($delete_entries)) if $debug_level > 0;
	delay_queue_delete( $delete_entries);
    } else {
	$logger->log(LOG_ERR, "0 entries be deleted") if $debug_level > 0;
    }
    if (%{$retry_entries}) {
	$logger->log(LOG_ERR, "entries to be retried:".Dumper($retry_entries)) if $debug_level > 0;
	delay_queue_retry( $retry_entries);
    } else {
	$logger->log(LOG_ERR, "0 entries to be retried") if $debug_level > 0;
    }
}

1;

