package PureMessage::Policy::DelayMail;

use strict;
use warnings;

use base 'PureMessage::Policy::Plugin';
use Math::Round;
use PureMessage::MessageFile;
use File::Copy;
use Sophos::Config qw(:path);
use PureMessage::Policy::Delay::Config qw(%config);
use PureMessage::Version qw($PREFIX);

our @EXPORT_OK = qw(pmx_suspect_delay pmx_delayed_mail);

my $last_id_file = "$PREFIX/var/qdir/delay/last_id";
my $dbconf_file = "$PREFIX/var/qdir/db.conf";
my $delay_queue_size_file = "$PREFIX/var/qdir/delay/delay_queue_size";
my $delay_dir = "$PREFIX/var/qdir/delay";
my $delay_dbentries_dir = "$delay_dir/db_entries";
my $delay_min_minutes = $config{delay_min_minutes} || 10;
my $delay_max_minutes = $config{delay_max_minutes} || 60;
my $max_msg_cmpnnt_len = 255;
my $delay_queue_status = 0;
my $location_id;
init_me();

sub init_me {
    my $db_conf = Sophos::Config->new;
    $db_conf->parse_file( $dbconf_file );
    $location_id = path_value($db_conf->root, "location_id");
    if ($delay_min_minutes < 2 || $delay_min_minutes > 60) {
	$delay_min_minutes = 10;
    }
    if ($delay_max_minutes < 2 || $delay_max_minutes > 60) {
	$delay_max_minutes = 60;
    }
    if ($delay_min_minutes > $delay_max_minutes) {
	$delay_min_minutes = 10;
	$delay_max_minutes = 60;
    }
    $delay_queue_status = ( lc ($config{delay_status})  eq  'on');
    return;
}

sub new {
    my $class = shift;
    my %opts  = @_;
    return bless \%opts, $class;
}

sub _get_header_value {
    my ($name, $header) = @_;
    return $1 if ($header =~ /^(?:.*,)?$name=([^,]*)/);
    return undef;
}


sub _clean_mail_addresses {
    my $v = shift;
    require Mail::Address;
    local $Mail::Address::VERBOSE = 0; # silence "Unmatched () ... " warnings
    my @addr = grep length, map lc($_->address), Mail::Address->parse($v);
    return @addr if wantarray;
    return $addr[0];
}


sub set_delay_queue_size {
    my ( $size_to_add, $messages_to_add ) = @_;
    my $delay_queue_size_conf = Sophos::Config->new;

    $delay_queue_size_conf->parse_file(
	$delay_queue_size_file,
	create   => 1,
	writable => 1
    );

    path_value(
	$delay_queue_size_conf->root, "disk_used",
	$size_to_add + path_value($delay_queue_size_conf->root, "disk_used")
    );

    path_value(
	$delay_queue_size_conf->root, "message_count",
	path_value($delay_queue_size_conf->root, "message_count") + $messages_to_add
    );

    $delay_queue_size_conf->save;
}

sub  do_db_file_entry {
    my $last_id = shift;
    my $dir_delay_minute = shift;
    my $delay_time = shift;
    my $filename = shift;
    my $msg = shift;
    my $global_id;
    my $message_size;
    my $date;
    my $h_from;
    my $e_from;
    my @h_to;
    my @env_rcpt;
    my @mapped_rcpt;
    my @cto_rcpt;
    my $addr;
    my $fh;
    my $size;

    $global_id = "$last_id-$location_id";
    $message_size = $msg->message_size();
    $date = time;
    $h_from = _clean_mail_addresses($msg->header("From")) || '';
    $e_from = $msg->minfo_value("Envelope-From") || '';
    my @header_to = $msg->header("To");

    # get header "TO" 
    for my $addr (@header_to) {
	 $addr = _clean_mail_addresses($addr);
	next if grep $_ eq $addr, @h_to;
	if (length($addr) > $max_msg_cmpnnt_len) {
	    substr($addr, $max_msg_cmpnnt_len) = "";
	}
	push(@h_to, $addr);
    }

    for( $h_from, $e_from) {
	next unless defined;
	# make sure they fit
	if (length($_) > $max_msg_cmpnnt_len) {
	    substr($_, $max_msg_cmpnnt_len) = "";
	}
    }

    # envelope recipients
    foreach $addr ($msg->minfo_value("Envelope-To")) {
	$addr = _clean_mail_addresses($addr);
        # make sure they fit
        if (length($addr) > $max_msg_cmpnnt_len) {
	    substr($addr, $max_msg_cmpnnt_len) = "";
        }
	push @env_rcpt , $addr;
    }

    # mapped recipients
    if (my @mrecip = $msg->minfo_value("Mapped-Recipients")) {
        foreach $addr (@mrecip) {
            my $addr = _clean_mail_addresses($addr);
            # make sure they fit
            if (length($addr) > $max_msg_cmpnnt_len) {
		substr($addr, $max_msg_cmpnnt_len) = "";
            }
	    push @mapped_rcpt, $addr;
        }
    }
	
    # c_to recipients BCC,CC
    for my $h (qw(To Cc Bcc Resent-To Resent-Cc Resent-Bcc)) {
        my @v = $msg->header($h);
        next unless @v;
        for my $addr (@v) {
	    $addr = _clean_mail_addresses($addr);
            next if grep $_ eq $addr, @cto_rcpt;
            if (length($addr) > $max_msg_cmpnnt_len) {
		substr($addr, $max_msg_cmpnnt_len) = "";
            }
            push(@cto_rcpt, $addr);
        }
    }

    $filename = "$delay_dbentries_dir/$filename.$$";
    open($fh, '>>', $filename ) or die "can't open file $filename for $!\n";
    print $fh "$global_id\n$dir_delay_minute\n$delay_time\n$message_size\n$date\n$h_from\n$e_from\n";

    $size = @h_to; 
    print $fh "$size\n";
    if ($size > 0) {
	$addr = join("\n", @h_to);
	print $fh "$addr\n";
    }

    $size = @env_rcpt; 
    print $fh "$size\n";
    if ($size > 0) {
	$addr = join("\n", @env_rcpt);
	print $fh "$addr\n";
    }
    $size = @mapped_rcpt; 
    print $fh "$size\n";
    if ($size > 0) {
	$addr = join("\n", @mapped_rcpt);
	print $fh "$addr\n";
    }
    $size = @cto_rcpt; 
    print $fh "$size\n";
    if ($size > 0) {
	$addr = join("\n", @cto_rcpt);
	print $fh "$addr\n";
    }

    print $fh "\n";
    close $fh;
}

sub delay_mail {
    my $msg = shift;
    my $opts = shift;
    my $filename;

    my $last_id_conf = Sophos::Config->new;
    $last_id_conf->parse_file(
	$last_id_file,
	create   => 1,
	writable => 1
    );

    my $last_id = path_value($last_id_conf->root, "last_id") || 0;
    ++$last_id;
    path_value($last_id_conf->root, "last_id", $last_id);
    $last_id_conf->save;

    $filename = (localtime(time))[1];
    my $dir_delay_minute = $filename + $opts->{delay_time};
    $dir_delay_minute = $dir_delay_minute - 60 if $dir_delay_minute >= 60;

    --$dir_delay_minute if $opts->{delay_time} == 60;
    $dir_delay_minute = 59 if $dir_delay_minute < 0;

    my $delay_message_file = "$delay_dir/$dir_delay_minute/$last_id";
    my $temp_file = delete ($opts->{tempfile})->filename;

    copy($temp_file, $delay_message_file);
    my $mf = PureMessage::MessageFile->make($delay_message_file, $opts->{minfo});
    unless ( $mf ) { return 0; }

    $msg->log("DEBUG", "delay file=$last_id, minute dir=$dir_delay_minute");

    # Calculate the delay size in file to /opt/pmx6/var/qdir/delay/delay_queue_size
    my $mail_size = -s $delay_message_file;
    set_delay_queue_size($mail_size, 1);

    # do entry in file for this mail.
    do_db_file_entry($last_id, $dir_delay_minute, $opts->{delay_time}, $filename, $msg);

}



sub pmx_suspect_delay {
    my ($msg) = @_;
    our @delay_times;
    my $i;

    if (! $delay_queue_status) {
	return;
    }
    my $delta = ($delay_max_minutes - $delay_min_minutes) / 4;
    @delay_times = ();
    $i = 0;
    while ($i < 5) {
	$delay_times[$i] = int($delay_min_minutes + ($i * $delta));
	$i++;
    }

    # dnt delay the mail if AS is not giving DS_SUSP level
    my $category;
    for ($msg->minfo_value("Spam-Hit")) {
        $category = $1 if (/^DQ_SUSP_([1-5])$/ && (!defined $category || $category < $1));
    }
    return if (!defined $category);

    # dont delay the mail if delay resion is not timeout
    my $header = $msg->header('X-Sophos-ReScanReason');
    if ($header) {
        my $reason = _get_header_value('rr', $header);
        return if ($reason && $reason ne 'timeout');
    }

    my $delay_time = $delay_times[$category - 1] || return;

    # check for timeout of mail for delay queue
    my @delays = ();
    $header = $msg->header('X-Sophos-DelayHistory');
    if ($header) {
        my $sum_delay_time = $delay_time;
        @delays = split('#', $header);
        for my $delay (@delays) {
            my $time = _get_header_value('dt', $delay);
            $sum_delay_time += $time if ($time);
        }
        return if ($sum_delay_time > $delay_times[4]);
    }

    #push delay time in header X-Sophos-DelayHistory
    push @delays, "sc=$category,dt=$delay_time";
    my $value = join('#', @delays);
    $msg->log('INFO', "DelayHistory: $value");
    if ($header) {
        $msg->header_replace('X-Sophos-DelayHistory', $value);
    } else {
        $msg->header_add('X-Sophos-DelayHistory', $value);
    }
    my @action = ('delay');
    $msg->set_attr('mark_action', \@action);
    my @type = ('SuspectSpam');
    $msg->set_attr('mark_type', \@type);
   
    delay_mail($msg, {
        tempfile => $msg->implode_to_tempfile(),
        minfo => $msg->minfo(),
        delay_time => $delay_time
    });

    $msg->delivery_action('delay', 'Suspect Spam');
    $msg->stop;

    return;
}

sub pmx_delayed_mail {
    my $msg = shift;
    if ($msg->get_attr('mark_dq_reinjected')) {
        return 1;
    }
    return 0;
}

1;
